@@ -11,11 +11,13 @@
 #include "dat.h"
 #include "fns.h"
 #include "patchlevel.h"
+#include <stdlib.h>
+#include <X11/keysym.h>
 
 char    *version[] = 
 {
     "9wm version 1.1, Copyright (c) 1994 David Hogan", 
-    "w9wm version 1.0, Copyright (c) 2000 Benjamin Drieu", 0,
+    "w9wm version 0.2, Copyright (c) 2000 Benjamin Drieu", 0,
 };
 
 Display         *dpy;
@@ -62,11 +64,12 @@ char    *fontlist[] = {
     0,
 };
 
+int
 main(argc, argv)
 int argc;
 char    *argv[];
 {
-    int i, status, background, do_exit, do_restart, dummy;
+    int i, background, do_exit, do_restart, dummy;
     unsigned long mask;
     XEvent ev;
     XGCValues gv;
@@ -200,7 +203,7 @@ char    *argv[];
     attr.cursor = arrow;
     attr.event_mask = SubstructureRedirectMask
         | SubstructureNotifyMask | ColormapChangeMask
-        | ButtonPressMask | ButtonReleaseMask | PropertyChangeMask;
+        | ButtonPressMask | ButtonReleaseMask | KeyPressMask | PropertyChangeMask;
     mask = CWCursor|CWEventMask;
     XChangeWindowAttributes(dpy, root, mask, &attr);
     XSync(dpy, False);
@@ -240,6 +243,18 @@ char    *argv[];
 #endif
                 fprintf(stderr, "9wm: unknown ev.type %d\n", ev.type);
             break;
+	case KeyRelease:
+	  ev.xkey.window = current->window;
+	  XSendEvent(dpy, current->window, False, NoEventMask, &ev);
+	  break;
+	case KeyPress:
+	  if (XLookupKeysym(&(ev.xkey),0) == XK_Tab)
+	    {
+	      activatenext();
+	    }
+	  ev.xkey.window = current->window;
+	  XSendEvent(dpy, current->window, False, NoEventMask, &ev);
+	  break;
         case ButtonPress:
             button(&ev.xbutton);
             break;
@@ -300,6 +315,22 @@ char    *argv[];
     }
 }
 
+void
+activatenext()
+{
+  Client * c;
+
+  for (c=current->next; c && c != current; c = ((c && c->next) ? c->next : clients))
+    if (c->virtual == virtual && 
+	c->parent != root)	/* paranoid */
+      {
+	active(c);
+	XMapRaised(dpy, c->parent);
+	return;
+      }
+
+}
+
 void
 usage()
 {
@@ -397,7 +428,7 @@ configurereq(e)
 XConfigureRequestEvent *e;
 {
     XWindowChanges wc;
-    XConfigureEvent ce;
+/*      XConfigureEvent ce; */
     Client *c;
 
     /* we don't set curtime as nothing here uses it */
@@ -455,7 +486,7 @@ XMapRequestEvent *e;
     curtime = CurrentTime;
     c = getclient(e->window, 0);
     if (c == 0 || c->window != e->window) {
-        fprintf(stderr, "9wm: bogus mapreq %x %x\n", c, e->window);
+        fprintf(stderr, "9wm: bogus mapreq %p %p\n", c, (void*)e->window);
         return;
     }
 
@@ -472,7 +503,7 @@ XMapRequestEvent *e;
     case NormalState:
         XMapRaised(dpy, c->parent);
         XMapWindow(dpy, c->window);
-        setstate(c, NormalState);
+        setstate9(c, NormalState);
         if (c->trans != None && current && c->trans == current->window)
                 active(c);
         break;
@@ -539,7 +570,7 @@ void
 destroy(w)
 Window w;
 {
-    Client *c, **l;
+    Client *c;
 
     curtime = CurrentTime;
     c = getclient(w, 0);
@@ -579,12 +610,12 @@ XClientMessageEvent *e;
                 hide(c);
         }
         else
-            fprintf(stderr, "9wm: WM_CHANGE_STATE: format %d data %d w 0x%x\n",
-                e->format, e->data.l[0], e->window);
+            fprintf(stderr, "9wm: WM_CHANGE_STATE: format %d data %d w 0p%p\n",
+                e->format, (int)e->data.l[0], (void*)e->window);
         return;
     }
     fprintf(stderr, "9wm: strange ClientMessage, type 0x%x window 0x%x\n",
-        e->message_type, e->window);
+        (int)e->message_type, (int)e->window);
 }
 
 void
